#include <stdio.h>
#include <math.h>

#define STEPS 27
#define RAD_SCALE 0x100000000
#define RAD_SCALE_HALF (RAD_SCALE / 2)
#define RAD_SCALE_QUARTER (RAD_SCALE / 4)
#define MAG_SCALE 2608131300
#define PI 3.1415926535897932384626433832795028841971l

int cordic[STEPS];

void getCosSin(unsigned int desiredPhaseRads, int* cos, int* sin) {
	long long i, q, accPhaseRads;
	int negate, l, k;
	if (desiredPhaseRads >= RAD_SCALE_HALF) {
		desiredPhaseRads -= RAD_SCALE_HALF;
		negate = 1;
	} else {
		negate = 0;
	}
	if (desiredPhaseRads >= RAD_SCALE_QUARTER) {
		i = 0;
		q = MAG_SCALE;
		accPhaseRads = RAD_SCALE_QUARTER;
	} else {
		i = MAG_SCALE;
		q = 0;
		accPhaseRads = 0;
	}
	for (l = 0, k = 0; l < STEPS; l++, k++) {
		long long phaseRads = cordic[l];
		long long tmpI = i;
		if (desiredPhaseRads < accPhaseRads) {
			i += q >> k;
			q -= tmpI >> k;
			accPhaseRads -= phaseRads;
		} else {
			i -= q >> k;
			q += tmpI >> k;
			accPhaseRads += phaseRads;
		}
	}

	if (i < 0) i = 0;
	if (q < 0) q = 0;
	if (i >= RAD_SCALE) i = RAD_SCALE - 1;
	if (q >= RAD_SCALE) q = RAD_SCALE - 1;
	i >>= 9;
	q >>= 9;
	if (negate) {
		*sin = (int) -q;
		*cos = (int) -i;
	} else {
		*sin = (int) q;
		*cos = (int) i;
	}
}

int getPerfectSin24(double desiredPhaseRads) {
	return (int) floor(sin(desiredPhaseRads) * ((double)0x7fffff) + 0.5l);
}

char* toBinary(int arg) {
	static char buf[100];
	int i, j;
	for (i = 29, j = 0; i >= 0; i--, j++) {
		if (arg & (1 << i)) buf[j] = '1'; else buf[j] = '0';
	}
	buf[j] = 0;
	return buf;
}

int main(int argc, char** argv) {
	long long i;
	double k;
	long long steps = RAD_SCALE;
	int plusError = 0;
	int minusError = 0;
	for (i = 0, k = 1.0l; i < STEPS; i++, k /= 2.0l) {
		cordic[i] = (int) floor(atan(k) * ((double)RAD_SCALE) / (2.0l * PI) + 0.0l);
		printf("when %lld => lookupValue <= \"%s\";\n", i, toBinary(cordic[i]));
	}
	{
		int cordicSin, cordicCos;
		getCosSin(0x12345, &cordicCos, &cordicSin);
		printf("sin: %06x\n", cordicSin);
	}
	for (i = 0; i < steps; i++) {
		double argFloat = 2 * PI / ((double) steps) * ((double) i);
		unsigned int argInt = (int) floor( ((double) RAD_SCALE) / ((double) steps) * ((double) i) + 0.5l);
		int perfectSin = getPerfectSin24(argFloat);
		int cordicSin, cordicCos, delta;
		getCosSin(argInt, &cordicCos, &cordicSin);
		delta = perfectSin - cordicSin;
		if ((i % 10000000) == 0) printf("testing %lld, minusError: %i, plusError: %i\n", i, minusError, plusError);
		if (delta == -1) minusError++;
		if (delta == 1) plusError++;
		if (delta > 1 || delta < -1) {
			printf("i: %i, sin: %06x, cordic: %06x, delta: %i\n", i, perfectSin, cordicSin, delta);
			return 1;
		}
	}
	return 0;
}
