-- Copyright (c) 2007 Frank Buss (fb@frank-buss.de)
-- See license.txt for license
-- Cellular Automaton "Rule 30" random generator with 64 bit internal bits,
-- 24 bit output every 24 clock ticks and an additional random bit every tick

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity RandomGenerator is
    port(
        clock	 : in std_logic;
		reset	 : in std_logic;
		value    : out unsigned(23 downto 0);
		bitValue : out std_logic
    );
end entity RandomGenerator;

architecture rtl of RandomGenerator is

    constant SEED: unsigned(63 downto 0) := x"9a97f300bf7244fa";
    signal cells: unsigned(63 downto 0) := SEED;
    signal newValue: unsigned(23 downto 0) := (others => '0');
    signal counter: natural range 0 to 24 := 0;

begin

    process(clock, reset)
    begin
        if rising_edge(clock) then
            if reset = '1' then
                cells <= SEED;
                counter <= 0;
                value <= (others => '0');
                bitValue <= '0';
            else
                cells <= rotate_left(cells, 1) xor (cells or rotate_right(cells, 1));
                newValue(counter) <= cells(32);
                bitValue <= cells(32);
                if counter = 23 then
                    counter <= 0;
                    value <= newValue;
                else
                    counter <= counter + 1;
                end if;
            end if;
        end if;
    end process;
end architecture rtl;
